#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <sys/time.h>
#include <signal.h>
#include <errno.h>
#include <ctype.h>
#include <fcntl.h>
#include <vector>
#include <string>
#include <map>
#include <set>

#include "metacube.h"
#include "parse.h"
#include "server.h"
#include "serverpool.h"
#include "input.h"
#include "state.pb.h"

#define STREAM_ID "stream"
#define STREAM_URL "http://gruessi.zrh.sesse.net:4013/"

using namespace std;

ServerPool *servers = NULL;
volatile bool hupped = false;

void hup(int ignored)
{
	hupped = true;
}

int create_server_socket(int port)
{
	int server_sock = socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP);
	if (server_sock == -1) {
		perror("socket");
		exit(1);
	}

	int one = 1;
	if (setsockopt(server_sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) == -1) {
		perror("setsockopt(SO_REUSEADDR)");
		exit(1);
	}

	// We want dual-stack sockets. (Sorry, OpenBSD and Windows XP...)
	int zero = 0;
	if (setsockopt(server_sock, IPPROTO_IPV6, IPV6_V6ONLY, &zero, sizeof(zero)) == -1) {
		perror("setsockopt(IPV6_V6ONLY)");
		exit(1);
	}

	// Set as non-blocking, so the acceptor thread can notice that we want to shut it down.
	if (ioctl(server_sock, FIONBIO, &one) == -1) {
		perror("ioctl(FIONBIO)");
		exit(1);
	}

	sockaddr_in6 addr;
	memset(&addr, 0, sizeof(addr));
	addr.sin6_family = AF_INET6;
	addr.sin6_port = htons(port);

	if (bind(server_sock, reinterpret_cast<sockaddr *>(&addr), sizeof(addr)) == -1) {
		perror("bind");
		exit(1);
	}

	if (listen(server_sock, 128) == -1) {
		perror("listen");
		exit(1);
	}

	return server_sock;
}

void *acceptor_thread_run(void *arg)
{
	int server_sock = int(intptr_t(arg));
	while (!hupped) {
		// Since we are non-blocking, we need to wait for the right state first.
		// Wait up to 50 ms, then check hupped.
		pollfd pfd;
		pfd.fd = server_sock;
		pfd.events = POLLIN;

		int nfds = poll(&pfd, 1, 50);
		if (nfds == 0 || (nfds == -1 && errno == EAGAIN)) {
			continue;
		}
		if (nfds == -1) {
			perror("poll");
			usleep(100000);
			continue;
		}

		sockaddr_in6 addr;
		socklen_t addrlen = sizeof(addr);

		// Get a new socket.
		int sock = accept(server_sock, reinterpret_cast<sockaddr *>(&addr), &addrlen);
		if (sock == -1 && errno == EINTR) {
			continue;
		}
		if (sock == -1) {
			perror("accept");
			usleep(100000);
			continue;
		}

		// Set the socket as nonblocking.
		int one = 1;
		if (ioctl(sock, FIONBIO, &one) == -1) {
			perror("FIONBIO");
			exit(1);
		}

		// Pick a server, round-robin, and hand over the socket to it.
		servers->add_client(sock);
	}
	return NULL;
}

struct StatsThreadParameters {
	string stats_file;
	int stats_interval;
};
		
void *stats_thread_run(void *arg)
{
	const StatsThreadParameters *parms = reinterpret_cast<StatsThreadParameters *>(arg);
	while (!hupped) {
		int fd;
		FILE *fp;
		time_t now;
		vector<ClientStats> client_stats;

		// Open a new, temporary file.
		char *filename = strdup((parms->stats_file + ".new.XXXXXX").c_str());
		fd = mkostemp(filename, O_WRONLY);
		if (fd == -1) {
			perror(filename);
			free(filename);
			goto sleep;
		}

		fp = fdopen(fd, "w");
		if (fp == NULL) {
			perror("fdopen");
			close(fd);
			unlink(filename);
			free(filename);
			goto sleep;
		}

		now = time(NULL);
		client_stats = servers->get_client_stats();
		for (size_t i = 0; i < client_stats.size(); ++i) {
			fprintf(fp, "%s %s %d %llu\n",
			        client_stats[i].remote_addr.c_str(),
				client_stats[i].stream_id.c_str(),
				int(now - client_stats[i].connect_time),
				(long long unsigned)(client_stats[i].bytes_sent));
		}
		if (fclose(fp) == EOF) {
			perror("fclose");
			unlink(filename);
			free(filename);
			goto sleep;
		}
		
		if (rename(filename, parms->stats_file.c_str()) == -1) {
			perror("rename");
			unlink(filename);
		}

sleep:
		int left_to_sleep = parms->stats_interval;
		do {
			left_to_sleep = sleep(left_to_sleep);
		} while (left_to_sleep > 0);
	}
	return NULL;
}

// Serialize the given state to a file descriptor, and return the (still open)
// descriptor.
int make_tempfile(const CubemapStateProto &state)
{
	char tmpl[] = "/tmp/cubemapstate.XXXXXX";
	int state_fd = mkstemp(tmpl);
	if (state_fd == -1) {
		perror("mkstemp");
		exit(1);
	}

	string serialized;
	state.SerializeToString(&serialized);

	const char *ptr = serialized.data();
	size_t to_write = serialized.size();
	while (to_write > 0) {
		ssize_t ret = write(state_fd, ptr, to_write);
		if (ret == -1) {
			perror("write");
			exit(1);
		}

		ptr += ret;
		to_write -= ret;
	}

	return state_fd;
}

// Read the state back from the file descriptor made by make_tempfile,
// and close it.
CubemapStateProto read_tempfile(int state_fd)
{
	if (lseek(state_fd, 0, SEEK_SET) == -1) {
		perror("lseek");
		exit(1);
	}

	string serialized;
	char buf[4096];
	for ( ;; ) {
		ssize_t ret = read(state_fd, buf, sizeof(buf));
		if (ret == -1) {
			perror("read");
			exit(1);
		}
		if (ret == 0) {
			// EOF.
			break;
		}

		serialized.append(string(buf, buf + ret));
	}

	close(state_fd);  // Implicitly deletes the file.

	CubemapStateProto state;
	if (!state.ParseFromString(serialized)) {
		fprintf(stderr, "PANIC: Failed deserialization of state.\n");
		exit(1);
	}

	return state;
}

int main(int argc, char **argv)
{
	fprintf(stderr, "\nCubemap starting.\n");

	struct timeval serialize_start;
	bool is_reexec = false;

	string config_filename = (argc == 1) ? "cubemap.config" : argv[1];
	vector<ConfigLine> config = parse_config(config_filename);

	int port = fetch_config_int(config, "port", 1, 65535, PARAMATER_MANDATORY);
	int num_servers = fetch_config_int(config, "num_servers", 1, 20000, PARAMATER_MANDATORY);  // Insanely high max limit.

	servers = new ServerPool(num_servers);

	int server_sock = -1, old_port = -1;
	set<string> deserialized_stream_ids;
	map<string, Input *> deserialized_inputs;
	if (argc == 4 && strcmp(argv[2], "-state") == 0) {
		is_reexec = true;

		fprintf(stderr, "Deserializing state from previous process... ");
		int state_fd = atoi(argv[3]);
		CubemapStateProto loaded_state = read_tempfile(state_fd);

		serialize_start.tv_sec = loaded_state.serialize_start_sec();
		serialize_start.tv_usec = loaded_state.serialize_start_usec();

		// Deserialize the streams.
		for (int i = 0; i < loaded_state.streams_size(); ++i) {
			servers->add_stream_from_serialized(loaded_state.streams(i));
			deserialized_stream_ids.insert(loaded_state.streams(i).stream_id());
		}

		// Put back the existing clients. It doesn't matter which server we
		// allocate them to, so just do round-robin.
		for (int i = 0; i < loaded_state.clients_size(); ++i) {
			servers->add_client_from_serialized(loaded_state.clients(i));
		}

		// Deserialize the inputs. Note that we don't actually add them to any state yet.
		for (int i = 0; i < loaded_state.inputs_size(); ++i) {
			deserialized_inputs.insert(make_pair(
				loaded_state.inputs(i).stream_id(),
				new Input(loaded_state.inputs(i))));
		} 

		// Deserialize the server socket.
		server_sock = loaded_state.server_sock();
		old_port = loaded_state.port();

		fprintf(stderr, "done.\n");
	}

	// Find all streams in the configuration file, and create them.
	set<string> expecting_stream_ids = deserialized_stream_ids;
	for (unsigned i = 0; i < config.size(); ++i) {
		if (config[i].keyword != "stream") {
			continue;
		}
		if (config[i].arguments.size() != 1) {
			fprintf(stderr, "ERROR: 'stream' takes exactly one argument\n");
			exit(1);
		}
		string stream_id = config[i].arguments[0];
		if (deserialized_stream_ids.count(stream_id) == 0) {
			servers->add_stream(stream_id);
		}
		expecting_stream_ids.erase(stream_id);
	}

	// Warn about any servers we've lost.
	// TODO: Make an option (delete=yes?) to actually shut down streams.
	for (set<string>::const_iterator stream_it = expecting_stream_ids.begin();
	     stream_it != expecting_stream_ids.end();
	     ++stream_it) {
		string stream_id = *stream_it;
		fprintf(stderr, "WARNING: stream '%s' disappeared from the configuration file.\n",
			stream_id.c_str());
		fprintf(stderr, "         It will not be deleted, but clients will not get any new inputs.\n");
		if (deserialized_inputs.count(stream_id) != 0) {
			delete deserialized_inputs[stream_id];
			deserialized_inputs.erase(stream_id);
		}
	}

	// Open a new server socket if we do not already have one, or if we changed ports.
	if (server_sock != -1 && port != old_port) {
		fprintf(stderr, "NOTE: Port changed from %d to %d; opening new socket.\n", old_port, port);
		close(server_sock);
		server_sock = -1;
	}
	if (server_sock == -1) {
		server_sock = create_server_socket(port);
	}

	// See if the user wants stats.
	string stats_file = fetch_config_string(config, "stats_file", PARAMETER_OPTIONAL);
	int stats_interval = fetch_config_int(config, "stats_interval", 1, INT_MAX, PARAMETER_OPTIONAL, -1);
	if (stats_interval != -1 && stats_file.empty()) {
		fprintf(stderr, "WARNING: 'stats_interval' given, but no 'stats_file'. No statistics will be written.\n");
	}

	servers->run();

	pthread_t acceptor_thread;
	pthread_create(&acceptor_thread, NULL, acceptor_thread_run, reinterpret_cast<void *>(server_sock));

	// Find all streams in the configuration file, and create inputs for them.
	vector<Input *> inputs;
	for (unsigned i = 0; i < config.size(); ++i) {
		if (config[i].keyword != "stream") {
			continue;
		}
		assert(config[i].arguments.size() == 1);
		string stream_id = config[i].arguments[0];

		if (config[i].parameters.count("src") == 0) {
			fprintf(stderr, "WARNING: stream '%s' has no src= attribute, clients will not get any data.\n",
				stream_id.c_str());
			continue;
		}

		string src = config[i].parameters["src"];
		Input *input = NULL;
		if (deserialized_inputs.count(stream_id) != 0) {
			input = deserialized_inputs[stream_id];
			if (input->get_url() != src) {
				fprintf(stderr, "INFO: Stream '%s' has changed URL from '%s' to '%s', restarting input.\n",
					stream_id.c_str(), input->get_url().c_str(), src.c_str());
				delete input;
				input = NULL;
			}
			deserialized_inputs.erase(stream_id);
		}
		if (input == NULL) {
			input = new Input(stream_id, src);
		}
		input->run();
		inputs.push_back(input);
	}

	// All deserialized inputs should now have been taken care of, one way or the other.
	assert(deserialized_inputs.empty());

	// Start writing statistics.
	pthread_t stats_thread;
	StatsThreadParameters stats_parameters;  // Must live for as long as the stats thread does.
	if (!stats_file.empty()) {
		stats_parameters.stats_file = stats_file;
		stats_parameters.stats_interval = stats_interval;
		pthread_create(&stats_thread, NULL, stats_thread_run, &stats_parameters);
	}

	signal(SIGHUP, hup);
	
	struct timeval server_start;
	gettimeofday(&server_start, NULL);
	if (is_reexec) {
		// Measure time from we started deserializing (below) to now, when basically everything
		// is up and running. This is, in other words, a conservative estimate of how long our
		// “glitch” period was, not counting of course reconnects if the configuration changed.
		double glitch_time = server_start.tv_sec - serialize_start.tv_sec +
			1e-6 * (server_start.tv_usec - serialize_start.tv_usec);
		fprintf(stderr, "Re-exec happened in approx. %.0f ms.\n", glitch_time * 1000.0);
	}

	while (!hupped) {
		usleep(100000);
	}

	// OK, we've been HUPed. Time to shut down everything, serialize, and re-exec.
	gettimeofday(&serialize_start, NULL);

	if (!stats_file.empty()) {
		if (pthread_join(stats_thread, NULL) == -1) {
			perror("pthread_join");
			exit(1);
		}
	}
	if (pthread_join(acceptor_thread, NULL) == -1) {
		perror("pthread_join");
		exit(1);
	}

	CubemapStateProto state;
	state.set_serialize_start_sec(serialize_start.tv_sec);
	state.set_serialize_start_usec(serialize_start.tv_usec);
	state.set_server_sock(server_sock);
	state.set_port(port);

	for (size_t i = 0; i < inputs.size(); ++i) {
		inputs[i]->stop();
		state.add_inputs()->MergeFrom(inputs[i]->serialize());
	}

	for (int i = 0; i < num_servers; ++i) {	
		servers->get_server(i)->stop();

		CubemapStateProto local_state = servers->get_server(i)->serialize();

		// The stream state should be identical between the servers, so we only store it once.
		if (i == 0) {
			state.mutable_streams()->MergeFrom(local_state.streams());
		}
		for (int j = 0; j < local_state.clients_size(); ++j) {
			state.add_clients()->MergeFrom(local_state.clients(j));
		}
	}
	delete servers;

	fprintf(stderr, "Serializing state and re-execing...\n");
	int state_fd = make_tempfile(state);

	char buf[16];
	sprintf(buf, "%d", state_fd);

	for ( ;; ) {
		execlp(argv[0], argv[0], config_filename.c_str(), "-state", buf, NULL);
		perror("execlp");
		fprintf(stderr, "PANIC: re-exec of %s failed. Waiting 0.2 seconds and trying again...\n", argv[0]);
		usleep(200000);
	}
}
